@@ -1,7 +1,7 @@
 /* Android activity for Gforth on Android
 
   Authors: Bernd Paysan, Anton Ertl
-  Copyright (C) 2013,2014,2015,2016,2017,2018,2019 Free Software Foundation, Inc.
+  Copyright (C) 2013,2014,2015,2016,2017,2018,2019,2020,2021 Free Software Foundation, Inc.
 
   This file is part of Gforth.
 
@@ -35,6 +35,7 @@ import android.content.pm.ActivityInfo;
 import android.content.pm.PackageManager;
 import android.content.res.Configuration;
 import android.content.res.AssetFileDescriptor;
+import android.media.AudioManager;
 import android.media.MediaPlayer;
 import android.media.MediaPlayer.OnVideoSizeChangedListener;
 import android.location.Location;
@@ -107,6 +108,7 @@ public class Gforth
     private SensorManager sensorManager;
     private ClipboardManager clipboardManager;
     private AlarmManager alarmManager;
+    private AudioManager audioManager;
     private ConnectivityManager connectivityManager;
     private InputMethodManager inputMethodManager;
     private BroadcastReceiver recKeepalive, recConnectivity;
@@ -153,7 +155,7 @@ public class Gforth
     public native void onEventNative(int type, Object event);
     public native void onEventNative(int type, int event);
     public native void callForth(long xt); // !! use long for 64 bits !!
-    public native void startForth(String libdir, String locale, String startfile);
+    public native void startForth(String libdir, String locale, String startfile, String filedirs[]);
 
     // own subclasses
     static class GforthView extends SurfaceView implements SurfaceHolder.Callback2 {
@@ -516,12 +518,13 @@ public class Gforth
 	sensorManager=(SensorManager)getSystemService(Context.SENSOR_SERVICE);
 	clipboardManager=(ClipboardManager)getSystemService(Context.CLIPBOARD_SERVICE);
 	alarmManager=(AlarmManager)getSystemService(Context.ALARM_SERVICE);
-	connectivityManager = (ConnectivityManager) getSystemService(Context.CONNECTIVITY_SERVICE);
-	inputMethodManager = (InputMethodManager)getSystemService(Context.INPUT_METHOD_SERVICE);
-	powerManager = (PowerManager)getSystemService(Context.POWER_SERVICE);
-	notificationManager = (NotificationManager)getSystemService(Context.NOTIFICATION_SERVICE);
+	audioManager=(AudioManager)getSystemService(Context.AUDIO_SERVICE);
+	connectivityManager=(ConnectivityManager)getSystemService(Context.CONNECTIVITY_SERVICE);
+	inputMethodManager=(InputMethodManager)getSystemService(Context.INPUT_METHOD_SERVICE);
+	powerManager=(PowerManager)getSystemService(Context.POWER_SERVICE);
+	notificationManager=(NotificationManager)getSystemService(Context.NOTIFICATION_SERVICE);
 	if (Build.VERSION.SDK_INT >= 26) {
-	    notificationChannel = new NotificationChannel("gnu.gforth.notifications", "Messages", NotificationManager.IMPORTANCE_DEFAULT);
+	    notificationChannel=new NotificationChannel("gnu.gforth.notifications", "Messages", NotificationManager.IMPORTANCE_DEFAULT);
 	    notificationChannel.enableLights(true);
 	    notificationChannel.setShowBadge(true);
 	    
@@ -636,7 +639,7 @@ public class Gforth
 	    };
 	registerReceiver(recKeepalive, new IntentFilter("gnu.gforth.keepalive") );
 	
-	pintent = PendingIntent.getBroadcast(this, 0, new Intent("gnu.gforth.keepalive"), 0);
+	pintent = PendingIntent.getBroadcast(this, 0, new Intent("gnu.gforth.keepalive"), PendingIntent.FLAG_MUTABLE);
 
 	// intent for notifications
 	gforthintent = PendingIntent.getActivity
@@ -645,7 +648,7 @@ public class Gforth
 	     .setAction("gnu.gforth.Gforth_n2o.MESSAGE")
 	     .setFlags(Intent.FLAG_ACTIVITY_REORDER_TO_FRONT |
 		       Intent.FLAG_ACTIVITY_SINGLE_TOP),
-	     PendingIntent.FLAG_UPDATE_CURRENT);
+	     PendingIntent.FLAG_UPDATE_CURRENT | PendingIntent.FLAG_MUTABLE);
 
 	// intent for network connectivity (!!use netlink socket!!)
 	recConnectivity = new BroadcastReceiver() {
@@ -671,9 +674,13 @@ public class Gforth
 
     public void doStart() {
 	if(!started) {
+	    String filedirs[] = {
+		Environment.getExternalStorageDirectory().toString(),
+		getExternalFilesDir(null).toString(),
+		getFilesDir().toString() };
 	    startForth(getApplicationInfo().nativeLibraryDir,
 		       Locale.getDefault().toString() + ".UTF-8",
-		       startfile);
+		       startfile, filedirs);
 	    started=true;
 	}
 	activated = -1;
