@@ -2,7 +2,7 @@
 
 
   Authors: Anton Ertl, Bernd Paysan, Jens Wilke, David Kühling
-  Copyright (C) 1995,1996,1997,1998,2000,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018,2019 Free Software Foundation, Inc.
+  Copyright (C) 1995,1996,1997,1998,2000,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018,2019,2020,2021 Free Software Foundation, Inc.
 
   This file is part of Gforth.
 
@@ -128,10 +128,15 @@ int ufileattr[6]= {
 jmp_buf throw_jmp_handler;
 #endif
 
-#if defined(DOUBLY_INDIRECT)
-#  define CFA(n)	({Cell _n = (n); ((Cell)(((_n & 0x4000) ? symbols : xts)+(_n&~0x4000UL)));})
+#ifdef NEW_CFA
+#define CFA_BYTE_OFFSET         (2*sizeof(Cell))
 #else
-#  define CFA(n)	((Cell)(symbols+((n)&~0x4000UL)))
+#define CFA_BYTE_OFFSET         0
+#endif
+#if defined(DOUBLY_INDIRECT)
+#  define CFA(n)	({Cell _n = (n); (((Cell)(((_n & 0x4000) ? symbols : xts)+(_n&~0x4000UL)))+((_n & 0x4000) ? CFA_BYTE_OFFSET : 0));})
+#else
+#  define CFA(n)	(((Cell)(symbols+((n)&~0x4000UL)))+CFA_BYTE_OFFSET)
 #endif
 
 #define maxaligned(n)	(typeof(n))((((Cell)n)+sizeof(Float)-1)&-sizeof(Float))
@@ -162,6 +167,10 @@ int optind = 1;
 #ifndef MAP_NORESERVE
 #define MAP_NORESERVE 0
 #endif
+int map_32bit=0; /* mmap option, can be set to MAP_32BIT with --map_32bit */
+#ifndef MAP_32BIT
+#define MAP_32BIT 0
+#endif
 #if defined(__CYGWIN__) && defined(__x86_64)
 #define MAP_NORESERVE 0
 #endif
@@ -195,6 +204,7 @@ static int nonrelocs = 0;
 #ifdef HAS_DEBUG
 int debug=0;
 int debug_mcheck=0;
+int debug_prim=1;
 # define debugp(x...) do { if (debug) fprintf(x); } while (0)
 #else
 # define perror(x...)
@@ -423,7 +433,7 @@ void gforth_relocate(Address sections[], Char *bitstrings[],
 		    compile_prim1(&image[i]);
 		  }
 #endif
-		} else {
+		} else if(debug_prim) {
 		  Char * dumpa = (Char*)&image[i];
 		  for(; dumpa < (Char*)&image[i+8]; dumpa++) {
 		    fprintf(stderr, "%02x ", *dumpa);
@@ -456,7 +466,7 @@ void gforth_relocate(Address sections[], Char *bitstrings[],
 		  MAKE_CF(image+i,symbols[groups[group]+tok]);
 		}
 #endif
-	      } else {
+	      } else if(debug_prim) {
 		Char * dumpa = (Char*)&image[i];
 		for(; dumpa < (Char*)&image[i+8]; dumpa++) {
 		  fprintf(stderr, "%02x ", *dumpa);
@@ -552,15 +562,13 @@ static void after_alloc(Address r, Cell size)
 #if defined(HAVE_MMAP)
 static Address alloc_mmap(Cell size)
 {
-  void *r;
+  void *r=MAP_FAILED;
+  static int dev_zero=-1;
 
-#if defined(MAP_ANON)
-  debugp(stderr,"try mmap(%p, $%lx, ..., MAP_ANON, ...); ", NULL, size);
-  r = mmap(0, size, PROT_EXEC|PROT_READ|PROT_WRITE, MAP_ANON|MAP_PRIVATE|map_noreserve, -1, 0);
-#else /* !defined(MAP_ANON) */
+#if !defined(MAP_ANON)
   /* Ultrix (at least) does not define MAP_FILE and MAP_PRIVATE (both are
      apparently defaults) */
-  static int dev_zero=-1;
+  int MAP_ANON=MAP_FILE;
 
   if (dev_zero == -1)
     dev_zero = open("/dev/zero", O_RDONLY);
@@ -568,11 +576,15 @@ static Address alloc_mmap(Cell size)
     r = MAP_FAILED;
     debugp(stderr, "open(\"/dev/zero\"...) failed (%s), no mmap; ", 
 	      strerror(errno));
-  } else {
-    debugp(stderr,"try mmap(%p, $%lx, ..., MAP_FILE, dev_zero, ...); ", NULL, size);
-    r=mmap(0, size, PROT_EXEC|PROT_READ|PROT_WRITE, MAP_FILE|MAP_PRIVATE|map_noreserve, dev_zero, 0);
+    after_alloc(r, size);
+    return r;
   }
 #endif /* !defined(MAP_ANON) */
+  debugp(stderr,"try mmap(%p, $%lx, ..., dev_zero, ...); ", NULL, size);
+  if (MAP_32BIT && map_32bit)
+    r=mmap(0, size, PROT_EXEC|PROT_READ|PROT_WRITE, MAP_ANON|MAP_PRIVATE|map_noreserve|MAP_32BIT, dev_zero, 0);
+  if (r==MAP_FAILED)
+    r=mmap(0, size, PROT_EXEC|PROT_READ|PROT_WRITE, MAP_ANON|MAP_PRIVATE|map_noreserve, dev_zero, 0);
   after_alloc(r, size);
   return r;  
 }
@@ -616,7 +628,7 @@ Address gforth_alloc(Cell size)
   Address r;
 
   r=alloc_mmap(size);
-  if (r!=(Address)MAP_FAILED) {
+  if (r!=MAP_FAILED) {
     debugp(stderr, "mmap($%lx) succeeds, address=%p\n", (long)size, r);
     return r;
   }
@@ -936,6 +948,7 @@ MAYBE_UNUSED static Label bsearch_next(Label key, Label *a, UCell n)
     return bsearch_next(key, a, mid+1);
 }
 
+#ifndef NO_DYNAMIC
 static int state_map(int state)
 {
   if (state==0) return STACK_CACHE_DEFAULT;
@@ -943,7 +956,6 @@ static int state_map(int state)
   return state;
 }
 
-#ifndef NO_DYNAMIC
 static void gforth_printprims()
 {
   unsigned i;
@@ -1724,11 +1736,11 @@ void compile_prim1(Cell *start)
     return;
   prim = (Label)*start;
   if (prim<((Label)(xts+DOER_MAX)) || prim>((Label)(xts+npriminfos))) {
-    debugp(stderr,"compile_prim encountered xt %p [%lx]\n", prim, (*(Cell*)prim-(Cell)labels));
-    *start = (Cell)((*(Cell*)prim-(Cell)labels)+(Cell)vm_prims);
+    debugp(stderr,"compile_prim encountered xt %p [%lx]\n", prim, (((Cell)CODE_ADDRESS(prim)-(Cell)labels)));
+    *start = (Cell)(((Cell)CODE_ADDRESS(prim)-(Cell)labels)+(Cell)vm_prims)+CFA_BYTE_OFFSET;
     return;
   } else {
-    *start = (Cell)(prim-((Label)xts)+((Label)vm_prims));
+    *start = (Cell)(prim-((Cell)xts)+((Cell)vm_prims));
     return;
   }
 #elif defined(INDIRECT_THREADED)
@@ -1750,11 +1762,11 @@ void compile_prim1(Cell *start)
       return;
     }
   }
-  prim_num = ((Xt)*start)-vm_prims;
+  prim_num = (((Xt)*start)-vm_prims)-CFA_OFFSET;
   if(prim_num >= npriminfos) {
     /* try search prim number in vm_prims */
     int step, i;
-    UCell inst = **(UCell**)start;
+    UCell inst = CODE_ADDRESS(*(UCell**)start);
     for(i=1; i<npriminfos; i*=2);
     i/=2;
     for(step=i/2; step>0; step/=2) {
@@ -1775,7 +1787,7 @@ void compile_prim1(Cell *start)
     // debugp(stderr,"optimize_rewrite(...,%d)\n",ninsts);
     ninsts=0;
     append_jump();
-    *start = *(Cell *)*start;
+    *start = (Cell)CODE_ADDRESS(*start);
     return;
   }    
   assert(ninsts<MAX_BB);
@@ -2055,9 +2067,9 @@ ImageHeader* gforth_loader(char* imagename, char* path)
     
     bases[i] = INSECTION(section.base);
     sizes[i] = section.dp-section.base;
-    sections[i] = alloc_mmap_guard(section.end);
+    sections[i] = alloc_mmap_guard(section.size);
     fseek(imagefile, -sizeof(SectionHeader), SEEK_CUR);
-    debugp(stderr, "section base=%p, dp=%p, end=%p\n", section.base, section.dp, section.end);
+    debugp(stderr, "section base=%p, dp=%p, size=%lx\n", section.base, section.dp, section.size);
     if(fread(sections[i], 1, sizes[i], imagefile) != sizes[i]) break;
   }
   gforth_relocate(sections, reloc_bits, sizes, bases, vm_prims);
@@ -2250,6 +2262,7 @@ int gforth_args(int argc, char ** argv, char ** path, char ** imagename)
       {"fp-stack-size", required_argument, NULL, 'f'},
       {"locals-stack-size", required_argument, NULL, 'l'},
       {"vm-commit", no_argument, &map_noreserve, 0},
+      {"map-32bit", no_argument, &map_32bit, 1},
       {"path", required_argument, NULL, 'p'},
       {"version", no_argument, NULL, 'v'},
       {"help", no_argument, NULL, 'h'},
@@ -2259,6 +2272,7 @@ int gforth_args(int argc, char ** argv, char ** path, char ** imagename)
       {"clear-dictionary", no_argument, &clear_dictionary, 1},
       {"debug", no_argument, &debug, 1},
       {"debug-mcheck", no_argument, &debug_mcheck, 1},
+      {"no-debug-prim", no_argument, &debug_prim, 0},
       {"diag", no_argument, NULL, 'D'},
       {"die-on-signal", no_argument, &die_on_signal, 1},
       {"ignore-async-signals", no_argument, &ignore_async_signals, 1},
@@ -2334,6 +2348,7 @@ Engine Options:\n\
   -i FILE, --image-file=FILE	    Use image FILE instead of `gforth.fi'\n\
   -l SIZE, --locals-stack-size=SIZE Specify locals stack size\n\
   -m SIZE, --dictionary-size=SIZE   Specify Forth dictionary size\n\
+  --map-32bit			    Try to put the dictionary in the first 2GB\n\
   --no-dynamic			    Use only statically compiled primitives\n\
   --no-offset-im		    Load image at normal position\n\
   --no-super			    No dynamically formed superinstructions\n\
@@ -2605,22 +2620,15 @@ Xt gforth_find(Char * name)
   return xt;
 }
 
-Cell* winch_addr=0;
-
-void gforth_setwinch()
-{
-  Xt winch_query=gforth_find((Char*)"winch?");
-  if(winch_query != 0) {
-    gforth_execute(winch_query);
-    winch_addr = (Cell*)*gforth_SP++;
-  }
-}
+Cell winch_addr=0;
 
 void gforth_bootmessage()
 {
   Xt bootmessage=gforth_find((Char*)"bootmessage");
   if(bootmessage != 0) {
     gforth_execute(bootmessage);
+  } else {
+    debugp(stderr, "Can't print bootmessage\n");
   }
 }
 
@@ -2647,7 +2655,6 @@ Cell gforth_main(int argc, char **argv, char **env)
   debugp(stderr, "Start returned %ld\n", retvalue);
 
   while(retvalue == -56) { /* throw-code for quit */
-    gforth_setwinch();
     gforth_bootmessage();
     retvalue = gforth_quit();
   }
