@@ -1,7 +1,7 @@
 /* Gforth support functions
 
   Authors: Bernd Paysan, Anton Ertl
-  Copyright (C) 1995,1996,1997,1998,2000,2003,2004,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018,2019 Free Software Foundation, Inc.
+  Copyright (C) 1995,1996,1997,1998,2000,2003,2004,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018,2019,2020,2021 Free Software Foundation, Inc.
 
   This file is part of Gforth.
 
@@ -517,6 +517,8 @@ Cell rename_file(Char *c_addr1, UCell u1, Char *c_addr2, UCell u2)
 
 struct Cellquad read_line(Char *c_addr, UCell u1, FILE *wfileid)
 {
+  /* like read-line; the additional u3 result is the total number of
+     chars consumed (including newline characters) */
   UCell u2, u3;
   Cell flag, wior;
   Cell c;
@@ -524,32 +526,39 @@ struct Cellquad read_line(Char *c_addr, UCell u1, FILE *wfileid)
 
   flag=-1;
   u3=0;
+  wior=0;
   if (u1>0)
     gf_regetc(wfileid);
   for(u2=0; u2<u1; u2++) {
     do{
       c = getc(wfileid);
-      if((wior=FILEIO(ferror(wfileid)))) clearerr(wfileid);
-    } while (wior == TOIOR(EINTR));
-    if(wior) break;
+    } while (c == EOF && ferror(wfileid)==EINTR);
+    if (c==EOF) {
+      int err=ferror(wfileid);
+      clearerr(wfileid);
+      wior=FILEIO(err);
+      flag=FLAG(u2!=0);
+      break;
+    }
     u3++;
     if (c=='\n') break;
     if (c=='\r') {
-      do {
-	c = getc(wfileid);
-	if((wior=FILEIO(ferror(wfileid)))) clearerr(wfileid);
-      } while(wior==TOIOR(EINTR));
-      if(wior) break;
+      do{
+        c = getc(wfileid);
+      } while (c == EOF && ferror(wfileid)==EINTR);
+      if (c==EOF) {
+        int err=ferror(wfileid);
+        clearerr(wfileid);
+        wior=FILEIO(err);
+        flag=-1;
+        break;
+      }
       if (c!='\n')
 	gf_ungetc(c,wfileid);
       else
 	u3++;
       break;
     }
-    if (c==EOF) {
-      flag=FLAG(u2!=0);
-      break;
-    }
     c_addr[u2] = (Char)c;
   }
   r.n1 = u2;
@@ -824,7 +833,10 @@ UCell gforth_dlopen(Char *c_addr, UCell u)
 #ifndef RTLD_GLOBAL
 #define RTLD_GLOBAL 0
 #endif
-  lib = (UCell)dlopen(file, RTLD_GLOBAL);
+#ifndef RTLD_LAZY
+#define RTLD_LAZY 0
+#endif
+  lib = (UCell)dlopen(file, RTLD_GLOBAL | RTLD_LAZY);
   free_l(file);
   if(lib) return lib;
   fprintf(stderr, "%s\n", dlerror());
@@ -836,6 +848,45 @@ UCell gforth_dlopen(Char *c_addr, UCell u)
   return 0;
 }
 
+UCell gforth_dlopen2(Char *c_addr, UCell u)
+{
+  char * file=tilde_cstr(c_addr, u);
+  UCell lib;
+#if defined(HAVE_LIBDL) || defined(HAVE_DLOPEN)
+#ifndef RTLD_GLOBAL
+#define RTLD_GLOBAL 0
+#endif
+#ifndef RTLD_LAZY
+#define RTLD_LAZY 0
+#endif
+  lib = (UCell)dlopen(file, RTLD_LAZY | RTLD_GLOBAL);
+  free_l(file);
+  if(lib) return lib;
+  fprintf(stderr, "%s\n", dlerror());
+#elif defined(_WIN32)
+  lib = (UCell) GetModuleHandle(file);
+  free_l(file);
+  if(lib) return lib;
+#endif
+  return 0;
+}
+
+UCell gforth_dlsym2(Char *c_addr1, UCell u1, UCell u2)
+{
+  char * string = cstr(c_addr1, u1);
+  UCell u3;
+#if defined(HAVE_LIBDL) || defined(HAVE_DLOPEN)
+  u3 = (UCell) dlsym((void*)u2,string);
+#elif defined(_WIN32)
+  u3 = (Cell) GetProcAddress((HMODULE)u2, string);
+#else
+#warning Define lib-sym!
+  u3 = 0;
+#endif
+  free_l(string);
+  return u3;
+}
+
 void gforth_dlclose(UCell lib)
 {
 #if defined(HAVE_LIBLTDL)
@@ -847,6 +898,15 @@ void gforth_dlclose(UCell lib)
 #endif
 }
 
+void gforth_dlclose2(UCell lib)
+{
+#if defined(HAVE_LIBDL) || defined(HAVE_DLOPEN)
+  dlclose((void*)lib);
+#elif defined(_WIN32)
+  FreeLibrary(lib);
+#endif
+}
+
 #endif /* !defined(STANDALONE) */
 
 
