@@ -2,4 +2,4 @@
 which sudo || alias sudo=eval
 
 git clone https://github.com/GeraldWodni/swig.git
-(cd swig && ./autogen.sh && ./configure && make && sudo make install)
+(cd swig && ./autogen.sh && ./configure "$@" && make && sudo make install)
